import pymysql
import json

with open('jugo.json', 'r') as f:
    settings = json.load(f)

conn = pymysql.connect(host=settings['host'],
                        user=settings['user'],
                        password=settings['password'],
                        db=settings['db'],
                        autocommit=True,
                        charset=settings['charset'])

cursor = conn.cursor()




import jugo.utils.connection as db
from jugo.utils.exceptions import ItemNotInFieldset

class DataObjects():
    def __init__ (self, model):
        self.model = model
        self.objects = []
        self.idx = -1

    def __iter__ (self):
        return self

    def __getitem__(self, i):
        return self.objects[i]

    def __len__(self):
        return len(self.objects)

    def __next__ (self):
        self.idx += 1

        try:
            return self.objects[self.idx]
        except IndexError:
            raise StopIteration()


    def values(self):
        return self.objects

    def get_fields(self):
        base_fields = self.model.__class__.__base__.__dict__
        own_fields = self.model.__class__.__dict__
        fields = {}

        for item in base_fields:
            if base_fields[item].__class__.__base__.__name__ == 'DataType':
                fields.update({item: base_fields[item]})

        for item in own_fields:
            if own_fields[item].__class__.__base__.__name__ == 'DataType':
                fields.update({item: own_fields[item]})

        return fields

    def update(self, **params):
        fields = self.get_fields()
        fields_keys = fields.keys()
 
        update_query = ''

        for param in params:
            if param == 'id':
                continue

            if param not in fields_keys:
                raise ItemNotInFieldset(param, self.model)

            update_query += f'`{param}` = {fields[param].pre_handling(params[param])}, '

        
        db.cursor.execute(f'UPDATE `{self.model.table_name}` SET {update_query[:-2]} WHERE `id` in ({",".join([str(item.id) for item in self.objects])})')


        

            
            


        



from jugo.data.DataObjects import DataObjects
from jugo.data.DataObject import DataObject
from jugo.models import types
from jugo.utils.exceptions import ItemNotInFieldset, UsedReservedVariable
import jugo.utils.connection as db
import pymysql

RESERVED_VARS = ('id', 'j_select', 'j_limit', 'j_order')

class Model:
    id = types.Integer(auto_increment=True)

    def __str__(self):
        return self.__class__.__name__

    def __init__(self):
        self.table_name = self.__str__().lower()
        self.cursor  = db.conn.cursor(pymysql.cursors.DictCursor)

        for field in self.get_fields(only_own=True).keys():
            if field in RESERVED_VARS:
                raise UsedReservedVariable(field, self)

    def get_fields(self, only_own=False, only_base=False):
        base_fields = self.__class__.__base__.__dict__
        own_fields = self.__class__.__dict__
        fields = {}

        if only_own: base_fields = []
        if only_base: own_fields = []
        
    
        for item in base_fields:
            if base_fields[item].__class__.__base__.__name__ == 'DataType':
                fields.update({item: base_fields[item]})

        for item in own_fields:
            if own_fields[item].__class__.__base__.__name__ == 'DataType':
                fields.update({item: own_fields[item]})

        return fields


    def create(self, **params):
        fields = self.get_fields()
        fields_keys = fields.keys()

        create_query = ''

        for param in params:
            if param not in fields_keys:
                raise ItemNotInFieldset(param, self)

            if param == 'id':
                continue

            create_query += f'`{param}` = {fields[param].pre_handling(params[param])}, '

        db.cursor.execute(f'INSERT INTO `{self.table_name}` SET {create_query[:-2]}')


    def objects(self, **params):
        fields = self.get_fields()
        fields_keys = fields.keys()

        select = []
        limit = None
        order_by = None

        if params.get('j_select'):
            select = params.pop('j_select')
            if not 'id' in select:
                select.append('id')

        if params.get('j_limit'):
            limit = params.pop('j_limit')
        if params.get('j_order'):
            order_by = params.pop('j_order')

        select_query = ','.join([f'`{item}`' for item in select])
        order_by_query = ''
        where_query = ''

        if order_by:
            order_by_query = 'ORDER BY '
            for item in order_by:
                if item[0] == '-':
                    order_by_query += f'`{item[1:]}` DESC, '
                else:
                    order_by_query += f'`{item}` ASC, '


        if params:
            where_query = 'WHERE '
            for index, item in enumerate(params):
                if item not in fields_keys:
                    raise ItemNotInFieldset(item, self)

                if index == len(params) - 1:
                    where_query += f'`{item}` = {fields[item].pre_handling(params[item])}'
                else:
                    where_query += f'`{item}` = {fields[item].pre_handling(params[item])} AND '


        self.cursor.execute(f'SELECT {select_query if select_query else "*"} FROM `{self.table_name}` {where_query} {order_by_query[:-2]} {"LIMIT %d" % limit if limit else ""}')
        data = self.cursor.fetchall()


        data_objects = DataObjects(self)
        for field in data:
            obj = DataObject(self.table_name)
            for param in field:
                setattr(obj, param, fields[param].post_handling(field[param]))
                setattr(obj, f'{param}__datatype__', fields[param])

            data_objects.objects.append(obj)

        return data_objects
            




import jugo.utils.connection as db

class DataObject():
    def __init__(self, table_name):
        self.__tablename__ = table_name

    def save(self, update_fields=None):
        fields = self.__dict__
        update_query = ''

        for param in fields:
            if fields[param].__class__.__base__.__name__ == 'DataType' or param == 'id' or param == '__tablename__' or (update_fields and param not in update_fields):
                continue

            update_query += f'`{param}` = {fields[f"{param}__datatype__"].pre_handling(fields[param])}, '


        db.cursor.execute(f'UPDATE `{self.__tablename__}` SET {update_query[:-2]} WHERE `id` = {fields["id"]}')

            



import json

class DataType():
    max_length = None
    default = None
    extra = ''
    unsigned = False

    # Handilng data before sql request
    def pre_handling(self, value):
        return value

    # Handling data after sql request
    def post_handling(self, value):
        return value


class Index():
    def __init__(self, *fields):
        self.fields = fields


class BigInteger(DataType):
    definition = 'BIGINT'

    def __init__(self, unsigned=False, default=None):
        self.unsigned = unsigned
        self.default = default

    definition_f = lambda self: 'BIGINT'
    default_f = lambda self: f'DEFAULT {self.default}' if self.default is not None else ''
    extra_f = lambda self: self.extra + ' UNSIGNED' if self.unsigned else ''


class Integer(DataType):
    definition = 'INT'

    def __init__(self, default=None, unsigned=False, auto_increment=False):
        self.default = default
        self.unsigned = unsigned
        self.auto_increment = auto_increment
        if auto_increment:
            self.extra = 'AUTO_INCREMENT'
        

    definition_f = lambda self: 'INT'
    default_f = lambda self: f'DEFAULT {self.default}' if self.default is not None else ''
    extra_f = lambda self: self.extra + ' PRIMARY KEY' if self.auto_increment else '' + ' UNSIGNED' if self.unsigned else ''


class MediumInteger(DataType):
    definition = 'MEDIUMINT'

    def __init__(self, unsigned=False, default=None):
        self.unsigned = unsigned
        self.default = default

    definition_f = lambda self: 'MEDIUMINT'
    default_f = lambda self: f'DEFAULT {self.default}' if self.default is not None else ''
    extra_f = lambda self: self.extra + ' UNSIGNED' if self.unsigned else ''


class SmallInteger(DataType):
    definition = 'SMALLINT'

    def __init__(self, unsigned=False, default=None):
        self.unsigned = unsigned
        self.default = default

    definition_f = lambda self: 'SMALLINT'
    default_f = lambda self: f'DEFAULT {self.default}' if self.default is not None else ''
    extra_f = lambda self: self.extra + ' UNSIGNED' if self.unsigned else ''


class TinyInteger(DataType):
    definition = 'TINYINT'

    def __init__(self, unsigned=False, default=None):
        self.unsigned = unsigned
        self.default = default

    definition_f = lambda self: 'TINYINT'
    default_f = lambda self: f'DEFAULT {self.default}' if self.default is not None else ''
    extra_f = lambda self: self.extra + ' UNSIGNED' if self.unsigned else ''


class Boolean(DataType):
    definition = 'TINYINT'

    def __init__(self, default=None):
        if default is not None:
            self.default = self.pre_handling(default)

    def pre_handling(self, value):
        return 1 if value else 0

    def post_handling(self, value):
        return True if value else False

    
    definition_f = lambda self: 'TINYINT(1)'
    default_f = lambda self: f'DEFAULT {self.default}' if self.default is not None else ''
    extra_f = lambda self: self.extra


class Char(DataType):
    definition = 'CHAR'
    def __init__(self, max_length=56, default=None):
        self.max_length = max_length
        self.default = default

    def pre_handling(self, value):
        return f'"{value}"'


    definition_f = lambda self: f'CHAR({self.max_length})'
    default_f = lambda self: f'DEFAULT "{self.default}"' if self.default else ''
    extra_f = lambda self: ''


class VarChar(DataType):
    definition = 'VARCHAR'
    def __init__(self, max_length=56, default=None):
        self.max_length = max_length
        self.default = default

    def pre_handling(self, value):
        return f'"{value}"'


    definition_f = lambda self: f'VARCHAR({self.max_length})'
    default_f = lambda self: f'DEFAULT "{self.default}"' if self.default else ''
    extra_f = lambda self: ''


class DateTime(DataType):
    definition = 'DATETIME'

    def __init__(self, auto_now=None, auto_now_add=None):
        if auto_now_add:
            self.default = 'CURRENT_TIMESTAMP'
        if auto_now:
            self.default = 'CURRENT_TIMESTAMP'
            self.extra = 'ON UPDATE CURRENT_TIMESTAMP'

    def pre_handling(self, value):
        return value.strftime('"%Y-%m-%d %H:%M:%S"')

    
    definition_f = lambda self: 'DATETIME'
    default_f = lambda self: f'DEFAULT {self.default}' if self.default else ''
    extra_f = lambda self: self.extra


class JSON(DataType):
    definition = 'TEXT'
    max_length = 65535

    def pre_handling(self, value):
        return f"'{json.dumps(value)}'"

    def post_handling(self, value):
        if value:
            return json.loads(value)
        else:
            return None

    
    definition_f = lambda self: 'TEXT'
    default_f = lambda self: ''
    extra_f = lambda self: self.extra

        
class LongText(DataType):
    definition = 'LONGTEXT'
    max_length = 4294967295

    definition_f = lambda self: 'LONGTEXT'
    default_f = lambda self: ''
    extra_f = lambda self: self.extra


class MediumText(DataType):
    definition = 'MEDIUMTEXT'
    max_length = 16777215

    definition_f = lambda self: 'MEDIUMTEXT'
    default_f = lambda self: ''
    extra_f = lambda self: self.extra


class Text(DataType):
    definition = 'TEXT'
    max_length = 65535

    definition_f = lambda self: 'TEXT'
    default_f = lambda self: ''
    extra_f = lambda self: self.extra


class TinyText(DataType):
    definition = 'TINYTEXT'
    max_length = 225

    definition_f = lambda self: 'TINYTEXT'
    default_f = lambda self: ''
    extra_f = lambda self: self.extra
    


    





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="Jugo",
    version="1.0.0",
    author="Vadim Shpak",
    author_email="lansi_bro@bk.ru",
    description="Tool for easy communicating with MySQL/MariaDB",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ThreadShakur/Jugo",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


import jugo.utils.connection as db


def get_fields(model):
    base_fields = model.__class__.__base__.__dict__
    own_fields = model.__class__.__dict__
    fields = {}

    for item in base_fields:
        if base_fields[item].__class__.__base__.__name__ == 'DataType':
            fields.update({item: base_fields[item]})

    for item in own_fields:
        if own_fields[item].__class__.__base__.__name__ == 'DataType':
            fields.update({item: own_fields[item]})

    return fields


def get_indexes(model):
    base_fields = model.__class__.__base__.__dict__
    own_fields = model.__class__.__dict__
    indexes = {}


    for item in base_fields:
        if base_fields[item].__class__.__name__ == 'Index':
            indexes.update({item: base_fields[item]})

    for item in own_fields:
        if own_fields[item].__class__.__name__ == 'Index':
            indexes.update({item: own_fields[item]})

    return indexes

def create_table(model):
    # Params of model
    fields = get_fields(model)
    fields_query = ''

    for item in fields:
        datatype = fields[item]

        fields_query += f'`{item}` {datatype.definition_f()} {datatype.default_f()} {datatype.extra_f()}, '

    db.cursor.execute(f'CREATE TABLE `{model.table_name}` ({fields_query[:-2]})  ENGINE=INNODB')


def check_for_changes(model):
    fields = get_fields(model)
    exists_in_table = []

    db.cursor.execute(f'SELECT `COLUMN_NAME`, `COLUMN_DEFAULT`, `DATA_TYPE`, `CHARACTER_MAXIMUM_LENGTH`, `EXTRA`, `COLUMN_KEY`, `COLUMN_TYPE` FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME="{model.table_name}"')
    table_fields = db.cursor.fetchall()

    for table_field in table_fields:
        column_finded = False

        # Iterating params of model, like name, date and etc.
        for param in fields:
            # Finding model param in table schema
            if param == table_field[0]:
                column_finded = True

                exists_in_table.append(param)
                datatype = fields[param]


                # table_field[1] = column default value
                # table_field[2] = column data type
                # table_field[3] = column value max_length
                # table_field[4] = column extra
                # table_field[5] = column key
                # table_field[6] = colunt type
                # If some of this params are not equals, updating table row

                if  (datatype.default != table_field[1] or
                    datatype.definition.lower() != table_field[2] or
                    datatype.max_length != table_field[3] or
                    datatype.extra.lower() != table_field[4].lower() or 
                    (datatype.unsigned and not 'unsigned' in table_field[6]) or
                    (not datatype.unsigned and 'unsigned' in table_field[6])):

                    if table_field[2] in ('int', 'tinyint') and str(datatype.default) == table_field[1]:
                        continue

                    print(f'ALTER field {model.table_name} {param}')

                    db.cursor.execute(f'ALTER TABLE `{model.table_name}` MODIFY COLUMN `{param}` {datatype.definition_f()} {datatype.default_f()} {datatype.extra_f()}')

                break

        # If field is not finded in our model schema - deleting it
        if not column_finded:
            print(f'DROP field {model.table_name} {table_field[0]}')
            db.cursor.execute(f'ALTER TABLE `{model.table_name}` DROP COLUMN `{table_field[0]}`;')

    for param in fields:
        # If field is not exist in table - creating it
        if param not in exists_in_table:
            datatype = fields[param]

            print(f'CREATE field {model.table_name} {param}')
            db.cursor.execute(f'ALTER TABLE `{model.table_name}` ADD `{param}` {datatype.definition_f()} {datatype.default_f()} {datatype.extra_f()}')

    # db.cursor.execute(f'SELECT `INDEX_NAME`, `COLUMN_NAME` FROM INFORMATION_SCHEMA.STATISTICS WHERE `TABLE_NAME` = "{model.table_name}" AND `TABLE_SCHEMA` = "{db.settings["db"]}"')
    # index_fields = db.cursor.fetchall()

    # indexes = get_indexes(model)
    # for index_field in index_fields:
    # for index in indexes:
    #     print(f'CREATE INDEX `{index}` ON `{model.table_name}` ({",".join([item for item in indexes[index].fields])});')
    #     db.cursor.execute(f'CREATE INDEX `{index}` ON `{model.table_name}` ({",".join([item for item in indexes[index].fields])});')


def syncdb(models):
    db.cursor.execute(f'SELECT `TABLE_NAME` FROM INFORMATION_SCHEMA.TABLES WHERE `TABLE_SCHEMA` = "{db.settings["db"]}"')

    tables = [item[0] for item in db.cursor.fetchall()]

    exist_tables = []
    models = __import__(models).__dict__

    for model in models:
        if '__call__' in models[model].__class__.__dict__.keys() and model != 'Model':
            item = models[model].__call__()

            if item.__class__.__base__.__name__ == 'Model':
                exist_tables.append(item.table_name)
                if item.table_name not in tables:
                    create_table(item)
                else:
                    check_for_changes(item)
    
    for table in tables:
        if table not in exist_tables:
            print(f'DROP table {table}')
            db.cursor.execute(f'DROP TABLE `{table}`')


                



class ItemNotInFieldset(Exception):
    def __init__(self, field, model):
        self.field = field
        self.model = model
    
    def __str__(self):
        return f"Unknown field {self.field} in {self.model}"

class UsedReservedVariable(Exception):
    def __init__(self, var, model):
        self.var = var
        self.model = model
    
    def __str__(self):
        return f"Variable {self.var} in {self.model} is RESERVED, please use another one"